//
//PDF Table Builder (using iText(R))
//written by Seiji Takegata (takegata@veltec.co.jp)
//

package jp.veltec.pdftable;

import java.util.HashMap;
import java.io.OutputStream;
import java.io.IOException;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;
import org.w3c.dom.NamedNodeMap;

import com.itextpdf.text.Document;
import com.itextpdf.text.DocumentException;
import com.itextpdf.text.pdf.PdfWriter;
import com.itextpdf.text.pdf.BaseFont;
import com.itextpdf.text.pdf.PdfContentByte;
import com.itextpdf.text.pdf.PdfPTable;
import com.itextpdf.text.pdf.PdfPCell;
import com.itextpdf.text.Font;
import com.itextpdf.text.Phrase;
import com.itextpdf.text.Paragraph;
import com.itextpdf.text.Chunk;
import com.itextpdf.text.Element;
import com.itextpdf.text.PageSize;
import com.itextpdf.text.Rectangle;
import com.itextpdf.text.BaseColor;

public class PdfTableBuilder{

	OutputStream fOutputStream;
	HashMap<String,String> fData;
	org.w3c.dom.Element fTemplate;
    Document fDocument;
	HashMap<String,String> fTableDefault;
	HashMap<String,String> fRowDefault;

	BaseFont bfMincho;
	BaseFont bfGothic;

	public PdfTableBuilder(){
		try{
			fDocument = new Document(PageSize.A4, getSize("15mm"), getSize("15mm"), getSize("15mm"), getSize("15mm"));
			bfMincho = BaseFont.createFont("HeiseiMin-W3","UniJIS-UCS2-H",BaseFont.NOT_EMBEDDED);
			bfGothic = BaseFont.createFont("HeiseiKakuGo-W5","UniJIS-UCS2-H",BaseFont.NOT_EMBEDDED);
		}catch(DocumentException e){
			e.printStackTrace();
		}catch(IOException e){
			e.printStackTrace();
		}
	}

	public void setOutputStream(OutputStream aOutputStream){
		fOutputStream = aOutputStream;
	}

	public void setData(HashMap<String,String> aData){
		fData = aData;
	}

	public void setTemplate(org.w3c.dom.Element aTemplate){
		fTemplate=aTemplate;
		String strSize = fTemplate.getAttribute("size");
		if(!strSize.equals("")){
			if(strSize.equals("A4P")){
				//same as default;
			}else if(strSize.equals("A4L")){
				fDocument.setPageSize(new Rectangle(getSize("297mm"),getSize("210mm")));
			}else if(strSize.equals("m0")){
				fDocument.setPageSize(new Rectangle(getSize("230mm"),getSize("120mm")));
			}else{
				String[] size = strSize.split(" ");
				if(size.length==2){
					fDocument.setPageSize(new Rectangle(getSize(size[0]),getSize(size[1])));
				}else{
					throw new IllegalArgumentException("not supported size:" + strSize);
				}
			}
		}

		String strMargin = fTemplate.getAttribute("margin");
		if(!strMargin.equals("")){
			String[] margines = strMargin.split(" ");
			if(margines.length!=4){
				throw new IllegalArgumentException("document margin setting must be: Lmm Rmm Tmm Bmm");
			}
			fDocument.setMargins(getSize(margines[0]),getSize(margines[1]),getSize(margines[2]),getSize(margines[3]));
		}

	}

    public void process(){
		try{
	        PdfWriter writer = PdfWriter.getInstance(fDocument, fOutputStream);
	        fDocument.open();
			NodeList tables = fTemplate.getElementsByTagName("table");
			for(int tableCount=0;tableCount<tables.getLength();tableCount++){
				org.w3c.dom.Element xmlTable = (org.w3c.dom.Element)tables.item(tableCount);
				String strCols = xmlTable.getAttribute("cols");
				if(strCols.equals("")){
					throw new IllegalStateException("cols not specified");
				}
				int cols = Integer.parseInt(strCols);
				PdfPTable pdfTable = new PdfPTable(cols);

				float outerBorderWidth = getSize("0.2mm");
				float innerBorderWidth = getSize("0.2mm");
				String attribute;
				attribute=xmlTable.getAttribute("outer-border-width");
				if(!attribute.equals("")){
					outerBorderWidth=getSize(attribute);
				}
				attribute=xmlTable.getAttribute("inner-border-width");
				if(!attribute.equals("")){
					innerBorderWidth=getSize(attribute);
				}
				attribute=xmlTable.getAttribute("margin-top");
				if(!attribute.equals("")){
					pdfTable.setSpacingBefore(getSize(attribute));
				}
				attribute=xmlTable.getAttribute("margin-bottom");
				if(!attribute.equals("")){
					pdfTable.setSpacingAfter(getSize(attribute));
				}
				attribute=xmlTable.getAttribute("widths");
				if(!attribute.equals("")){
					String[] strWidths = attribute.split(" ");
					int[] intWidths = new int[strWidths.length];
					for(int j=0;j<strWidths.length;j++){
						intWidths[j]=Integer.parseInt(strWidths[j]);
					}
					pdfTable.setWidths(intWidths);
				}
				fTableDefault = new HashMap<String,String>();
				NamedNodeMap mapT = xmlTable.getAttributes();
				for(int i=0;i<mapT.getLength();i++){
					Node node = mapT.item(i);
					fTableDefault.put(node.getNodeName(),node.getNodeValue());
				}

				NodeList trs = xmlTable.getElementsByTagName("tr");
				int rows=trs.getLength();
				boolean[][] matrix = new boolean[rows][cols];
				boolean isTopEnd=true;
				for(int rowCounter=0;rowCounter<rows;rowCounter++){
					org.w3c.dom.Element tr = (org.w3c.dom.Element)trs.item(rowCounter);
					fRowDefault = new HashMap<String,String>();
					NamedNodeMap mapR = tr.getAttributes();
					for(int i=0;i<mapR.getLength();i++){
						Node nodeR = mapR.item(i);
						fRowDefault.put(nodeR.getNodeName(),nodeR.getNodeValue());
					}
					NodeList tds = tr.getElementsByTagName("td");
					boolean isLeftEnd;
					boolean isRightEnd;
					boolean isBottomEnd;
					for(int k=0;k<tds.getLength();k++){
						org.w3c.dom.Element td = (org.w3c.dom.Element)tds.item(k);
						String content = td.getTextContent();
						String id = td.getAttribute("id");
						if(!id.equals("") && fData.get(id)!=null){
							content = fData.get(id);
						}

						Cell cell = new Cell();
						String strColspan = td.getAttribute("colspan");
						int colspan = 1;
						if(!strColspan.equals("")){
							colspan = Integer.parseInt(strColspan);
							cell.setColspan(colspan);
						}
						int colCounter=0;
						for(colCounter=0;colCounter<cols;colCounter++){
							if(matrix[rowCounter][colCounter]==false){
								break;
							}
						}
						if(colCounter==0){
							isLeftEnd=true;
						}else{
							isLeftEnd=false;
						}
						if(colCounter+colspan==cols){
							isRightEnd=true;
						}else{
							isRightEnd=false;
						}
						int rowspan = 1;
						String strRowspan = td.getAttribute("rowspan");
						if(!strRowspan.equals("")){
							rowspan = Integer.parseInt(strRowspan);
							cell.setRowspan(rowspan);
						}
						for(int r=0;r<rowspan;r++){
							for(int c=0;c<colspan;c++){
								matrix[rowCounter+r][colCounter+c]=true;
							}
						}
						if(rowCounter+rowspan==rows){
							isBottomEnd=true;
						}else{
							isBottomEnd=false;
						}
						cell.setBorderWidthLeft(innerBorderWidth);
						cell.setBorderWidthRight(innerBorderWidth);
						cell.setBorderWidthTop(innerBorderWidth);
						cell.setBorderWidthBottom(innerBorderWidth);
						if(isTopEnd){
							cell.setBorderWidthTop(outerBorderWidth);
						}
						if(isLeftEnd){
							cell.setBorderWidthLeft(outerBorderWidth);
						}
						if(isRightEnd){
							cell.setBorderWidthRight(outerBorderWidth);
						}
						if(isBottomEnd){
							cell.setBorderWidthBottom(outerBorderWidth);
						}

						cell.setAttributes(td);
						cell.setContent(content);
						pdfTable.addCell(cell);
					}
					
					isTopEnd=false;
				}
				fDocument.add(pdfTable);
			}

	        fDocument.close();
		}catch(DocumentException e){
			e.printStackTrace();
		}
    }

	class Cell extends PdfPCell{
		HashMap<String,String> attributes = new HashMap<String,String>();
		Font fFont  = new Font(bfMincho, 12);
		float fFontSize = 12f;
		float fLineHeight=1f;
		float fCharacterSpacing=0f;
		void setAttributes(org.w3c.dom.Element td){
			String attribute;
			attributes.putAll(fTableDefault);
			attributes.putAll(fRowDefault);
			NamedNodeMap map = td.getAttributes();
			for(int i=0;i<map.getLength();i++){
				Node node = map.item(i);
				attributes.put(node.getNodeName(),node.getNodeValue());
			}
			attribute = attributes.get("min-height");
			if(attribute!=null){
				setMinimumHeight(getSize(attribute));
			}
			attribute = attributes.get("height");
			if(attribute!=null){
				setFixedHeight(getSize(attribute));
			}
			attribute = attributes.get("align-h");
			if(attribute!=null){
				if(attribute.equals("center")){
					setHorizontalAlignment(Element.ALIGN_CENTER);
				}else if(attribute.equals("right")){
					setHorizontalAlignment(Element.ALIGN_RIGHT);
				}else if(attribute.equals("left")){
					setHorizontalAlignment(Element.ALIGN_LEFT);
				}
			}
			attribute = attributes.get("align-v");
			if(attribute!=null){
				if(attribute.equals("middle")){
					setVerticalAlignment(Element.ALIGN_MIDDLE);
				}else if(attribute.equals("bottom")){
					setVerticalAlignment(Element.ALIGN_BOTTOM);
				}else if(attribute.equals("top")){
					setVerticalAlignment(Element.ALIGN_TOP);
				}
			}
			attribute = attributes.get("font-family");
			if(attribute!=null){
				if(attribute.equals("SVbN")){
					fFont = new Font(bfGothic,fFontSize);
				}else if(attribute.equals("")){
					fFont = new Font(bfMincho,fFontSize);
				}else{
					throw new IllegalArgumentException("not supported font-family:" + attribute);
				}
			}
			attribute = attributes.get("font-style");
			if(attribute!=null){
				fFont.setStyle(attribute);
			}
			attribute = attributes.get("font-size");
			if(attribute!=null){
				fFontSize=getSize(attribute);
				fFont.setSize(fFontSize);
			}
			attribute=attributes.get("font-color");
			if(attribute!=null){
				fFont.setColor(getColor(attribute));
			}
			attribute=attributes.get("background-color");
			if(attribute!=null){
				setBackgroundColor(getColor(attribute));
			}
			attribute = attributes.get("border-width");
			if(attribute!=null){
				//setBorderWidth(getSize(attribute));
				setBorderWidthTop(getSize(attribute));
				setBorderWidthBottom(getSize(attribute));
				setBorderWidthLeft(getSize(attribute));
				setBorderWidthRight(getSize(attribute));
			}
			attribute = attributes.get("border-width-top");
			if(attribute!=null){
				setBorderWidthTop(getSize(attribute));
			}
			attribute = attributes.get("border-width-bottom");
			if(attribute!=null){
				setBorderWidthBottom(getSize(attribute));
			}
			attribute = attributes.get("border-width-left");
			if(attribute!=null){
				setBorderWidthLeft(getSize(attribute));
			}
			attribute = attributes.get("border-width-right");
			if(attribute!=null){
				setBorderWidthRight(getSize(attribute));
			}
			attribute = attributes.get("border-color");
			if(attribute!=null){
				setBorderColor(getColor(attribute));
			}
			attribute = attributes.get("border-color-left");
			if(attribute!=null){
				setBorderColorLeft(getColor(attribute));
			}
			attribute = attributes.get("border-color-right");
			if(attribute!=null){
				setBorderColorRight(getColor(attribute));
			}
			attribute = attributes.get("border-color-top");
			if(attribute!=null){
				setBorderColorTop(getColor(attribute));
			}
			attribute = attributes.get("border-color-bottom");
			if(attribute!=null){
				setBorderColorBottom(getColor(attribute));
			}
			attribute = attributes.get("padding");
			if(attribute!=null){
				setPadding(getSize(attribute));
			}
			attribute = attributes.get("padding-left");
			if(attribute!=null){
				setPaddingLeft(getSize(attribute));
			}
			attribute = attributes.get("padding-right");
			if(attribute!=null){
				setPaddingRight(getSize(attribute));
			}
			attribute = attributes.get("padding-top");
			if(attribute!=null){
				setPaddingTop(getSize(attribute));
			}
			attribute = attributes.get("padding-bottom");
			if(attribute!=null){
				setPaddingBottom(getSize(attribute));
			}
			attribute = attributes.get("rotation");
			if(attribute!=null){
				if(attribute.equals("left")){
					setRotation(90);
				}else if(attribute.equals("down")){
					setRotation(180);
				}else if(attribute.equals("right")){
					setRotation(270);
				}
			}
			attribute = attributes.get("line-height");
			if(attribute!=null){
				setLeading(0,Float.parseFloat(attribute));
			}
			attribute = attributes.get("character-spacing");
			if(attribute!=null){
				fCharacterSpacing=getSize(attribute);
			}
		}

		void setContent(String content){
			content=content.replaceAll("\\\\n","\n");
			Paragraph p = new Paragraph(content,fFont);
			setPhrase(p);
		}
	}

	float getSize(String size) throws IllegalArgumentException{
		if(size.endsWith("mm")){
			size=size.substring(0,size.length()-2);
			return 72f/25.4f*(Float.parseFloat(size));
		}else if(size.endsWith("pt")){
			size=size.substring(0,size.length()-2);
			return Float.parseFloat(size);
		}else{
			throw new IllegalArgumentException("size format error:"+ size);
		}
	}

	BaseColor getColor(String color){
		if(color.equals("white")){
			color="#FFFFFF";
		}else if(color.equals("black")){
			color="#000000";
		}else if(color.equals("gray")){
			color="#808080";
		}else if(color.equals("red")){
			color="#FF0000";
		}else if(color.equals("green")){
			color="#00FF00";
		}else if(color.equals("blue")){
			color="#0000FF";
		}else if(color.equals("yellow")){
			color="#FFFF00";
		}
		try{
			int r =Integer.valueOf(color.substring(1,3),16);
			int g =Integer.valueOf(color.substring(3,5),16);
			int b =Integer.valueOf(color.substring(5,7),16);
			return new BaseColor(r,g,b);
		}catch(NumberFormatException e){
			throw new IllegalArgumentException("not supported color:" + color);
		}
	}

}
